@@ -27,7 +27,8 @@
 using System;
 using System.Collections;
 using System.IO;
-using Mono.Data.SqliteClient;
+using System.Threading;
+using Mono.Data.Sqlite;
 
 namespace Beagle.Util {
 
@@ -50,7 +51,7 @@ namespace Beagle.Util {
 			public Tag[]  Tags;
 		}
 
-
+		static private bool tried_connection;
 		static private SqliteConnection connection;
 		static Hashtable tagCache = null;
 		static Hashtable directoryCache = null;
@@ -65,10 +66,37 @@ namespace Beagle.Util {
 
 		static private SqliteConnection PhotoStoreConnection {
 			get {
-				if (connection == null && File.Exists (PhotoStorePath)) {
+				if (! tried_connection && File.Exists (PhotoStorePath)) {
 					connection = new SqliteConnection ();
-					connection.ConnectionString = "URI=file:" + PhotoStorePath;
-					connection.Open ();
+					connection.ConnectionString = "version=3,URI=file:" + PhotoStorePath;
+
+					// Try to open the f-spot store.  This
+					// will fail if there is a version
+					// mismatch.
+					try {
+						connection.Open ();
+
+						// Run a dummy SELECT statement to catch more errors
+						// indicating sqlite version mismatches.
+						using (SqliteCommand command = new SqliteCommand ()) {
+							command.Connection = connection;
+							command.CommandText = "SELECT id FROM tags WHERE name='dummy'";
+							SqliteDataReader reader = null;
+							while (reader == null) {
+								try {
+									reader = command.ExecuteReader ();
+								} catch (SqliteBusyException) {
+									Thread.Sleep (50);
+								}
+							}
+							reader.Close ();
+						}
+					} catch (ApplicationException) {
+						Logger.Log.Warn ("Unable to open F-Spot database: no sqlite3 database found");
+						connection = null;
+					}
+
+					tried_connection = true;
 				}
 			
 				return connection;
