@@ -124,13 +124,16 @@ namespace Beagle.Util {
 
 				return storage_dir;
 			}
+			set { lock (storage_dir_lock) { storage_dir = value; } }
 		}
 
 		static string remote_storage_dir;
 		static public string GetRemoteStorageDir (bool create)
 		{
 			if (remote_storage_dir == null) {
-				if ((! SystemInformation.IsPathOnBlockDevice (PathFinder.StorageDir) && Conf.Daemon.IndexSynchronization) ||
+				bool index_synchronization = Conf.Daemon.GetOption (Conf.Names.IndexSynchronization, true);
+
+				if ((! SystemInformation.IsPathOnBlockDevice (PathFinder.StorageDir) && index_synchronization) ||
 				    Environment.GetEnvironmentVariable ("BEAGLE_SYNCHRONIZE_LOCALLY") != null) {
 					string index_pointer = Path.Combine (StorageDir, "remote_storage_dir");
 
