@@ -1,7 +1,7 @@
 //
 // LuceneQueryingDriver.cs
 //
-// Copyright (C) 2004-2005 Novell, Inc.
+// Copyright (C) 2004-2007 Novell, Inc.
 //
 
 //
@@ -26,6 +26,7 @@
 
 using System;
 using System.Collections;
+using System.Collections.Generic;
 using System.Diagnostics;
 using System.Globalization;
 using System.IO;
@@ -48,7 +49,7 @@ namespace Beagle.Daemon {
 
 	public class LuceneQueryingDriver : LuceneCommon {
 
-		static public bool Debug = false;
+		static private bool Debug = false;
 
 		public delegate bool UriFilter (Uri uri);
 		public delegate double RelevancyMultiplier (Hit hit);
@@ -76,9 +77,6 @@ namespace Beagle.Daemon {
 				text_cache = TextCache.UserCache;
 		}
 
-		////////////////////////////////////////////////////////////////
-
-
 		////////////////////////////////////////////////////////////////
 
 		public Uri[] PropertyQuery (Property prop)
@@ -87,26 +85,50 @@ namespace Beagle.Daemon {
 			// index as well?
 
 			IndexReader primary_reader;
-			LNS.IndexSearcher primary_searcher;
-
 			primary_reader = LuceneCommon.GetReader (PrimaryStore);
-			primary_searcher = new LNS.IndexSearcher (primary_reader);
 
-			Term term = new Term (PropertyToFieldName (prop.Type, prop.Key), prop.Value);
-			LNS.TermQuery query = new LNS.TermQuery (term);
-			LNS.Hits hits = primary_searcher.Search (query);
+			Term term;
+			term = new Term (PropertyToFieldName (prop.Type, prop.Key), prop.Value.ToLower ());
 
-			Uri[] uri_list = new Uri [hits.Length ()];
-			for (int i = 0; i < hits.Length (); i++) {
-				Document doc;
-				doc = hits.Doc (i);
-				uri_list [i] = GetUriFromDocument (doc);
+			TermDocs term_docs;
+			term_docs = primary_reader.TermDocs ();
+			term_docs.Seek (term);
+
+			ArrayList uri_list = new ArrayList ();
+			while (term_docs.Next ()) {
+				Document doc = primary_reader.Document (term_docs.Doc ());
+				uri_list.Add (GetUriFromDocument (doc));
 			}
 
-			primary_searcher.Close ();
-			ReleaseReader (primary_reader);
+			term_docs.Close ();
+			LuceneCommon.ReleaseReader (primary_reader);
 
-			return uri_list;
+			return (Uri[]) uri_list.ToArray (typeof (Uri));
+		}
+
+		////////////////////////////////////////////////////////////////
+
+		public bool HasUri (Uri uri)
+		{
+			IndexReader primary_reader;
+			primary_reader = LuceneCommon.GetReader (PrimaryStore);
+
+			Term term;
+			term = new Term ("Uri", UriFu.UriToEscapedString (uri));
+
+			TermDocs term_docs;
+			term_docs = primary_reader.TermDocs ();
+			term_docs.Seek (term);
+
+			bool has_uri = false;
+
+			if (term_docs.Next ())
+				has_uri = true;
+
+			term_docs.Close ();
+			LuceneCommon.ReleaseReader (primary_reader);
+
+			return has_uri;
 		}
 
 		////////////////////////////////////////////////////////////////
@@ -122,9 +144,18 @@ namespace Beagle.Daemon {
 			if (Debug)
 				Logger.Log.Debug ("###### {0}: Starting low-level queries", IndexName);
 
-			Stopwatch sw;
-			sw = new Stopwatch ();
-			sw.Start ();
+			Stopwatch total, a, b, c, d, e, f;
+
+			total = new Stopwatch ();
+			a = new Stopwatch ();
+			b = new Stopwatch ();
+			c = new Stopwatch ();
+			d = new Stopwatch ();
+			e = new Stopwatch ();
+			f = new Stopwatch ();
+
+			total.Start ();
+			a.Start ();
 
 			// Assemble all of the parts into a bunch of Lucene queries
 
@@ -191,9 +222,15 @@ namespace Beagle.Daemon {
 				}
 			}
 
+			a.Stop ();
+			if (Debug)
+				Log.Debug ("###### {0}: Building queries took {1}", IndexName, a);
+
 			// If we have no required parts, give up.
 			if (primary_required_part_queries == null)
 				return;
+
+			b.Start ();
 			
 			//
 			// Now that we have all of these nice queries, let's execute them!
@@ -220,9 +257,15 @@ namespace Beagle.Daemon {
 			if (secondary_reader != null)
 				secondary_searcher = new LNS.IndexSearcher (secondary_reader);
 
+			b.Stop ();
+			if (Debug)
+				Log.Debug ("###### {0}: Readers/searchers built in {1}", IndexName, b);
 
-			// Possibly create our whitelists from the search subset.
+			// Build whitelists and blacklists for search subsets.
+			c.Start ();
 			
+			// Possibly create our whitelists from the search subset.
+
 			LuceneBitArray primary_whitelist = null;
 			LuceneBitArray secondary_whitelist = null;
 			
@@ -280,6 +323,13 @@ namespace Beagle.Daemon {
 				}
 			}
 
+			c.Stop ();
+			if (Debug)
+				Log.Debug ("###### {0}: Whitelists and blacklists built in {1}", IndexName, c);
+
+			// Now run the low level queries against our indexes.
+			d.Start ();
+
 			BetterBitArray primary_matches = null;
 
 			if (primary_required_part_queries != null) {
@@ -298,12 +348,11 @@ namespace Beagle.Daemon {
 
 			} 
 
-			sw.Stop ();
+			d.Stop ();
 			if (Debug)
-				Logger.Log.Debug ("###### {0}: Finished low-level queries in {1}", IndexName, sw);
-			sw.Reset ();
-			sw.Start ();
+				Logger.Log.Debug ("###### {0}: Low-level queries finished in {1}", IndexName, d);
 
+			e.Start ();
 			// Only generate results if we got some matches
 			if (primary_matches != null && primary_matches.ContainsTrue ()) {
 				GenerateQueryResults (primary_reader,
@@ -318,9 +367,16 @@ namespace Beagle.Daemon {
 						      IndexName);
 			}
 
+			e.Stop ();
+
+			if (Debug)
+				Log.Debug ("###### {0}: Query results generated in {1}", IndexName, e);
+
 			//
 			// Finally, we clean up after ourselves.
 			//
+
+			f.Start ();
 			
 			primary_searcher.Close ();
 			if (secondary_searcher != null)
@@ -329,10 +385,24 @@ namespace Beagle.Daemon {
 			if (secondary_reader != null)
 				ReleaseReader (secondary_reader);
 
-
-			sw.Stop ();
+			f.Stop ();
+			
 			if (Debug)
-				Logger.Log.Debug ("###### {0}: Processed query in {1}", IndexName, sw);
+				Log.Debug ("###### {0}: Readers/searchers released in {1}", IndexName, f);
+
+			total.Stop ();
+			if (Debug) {
+				Log.Debug ("###### {0}: Query time breakdown:", IndexName);
+				Log.Debug ("###### {0}:    Build queries {1,6} ({2:0.0}%)", IndexName, a, 100 * a.ElapsedTime / total.ElapsedTime);
+				Log.Debug ("###### {0}:      Got readers {1,6} ({2:0.0}%)", IndexName, b, 100 * b.ElapsedTime / total.ElapsedTime);
+				Log.Debug ("###### {0}:       Whitelists {1,6} ({2:0.0}%)", IndexName, c, 100 * c.ElapsedTime / total.ElapsedTime);
+				Log.Debug ("###### {0}:          Queries {1,6} ({2:0.0}%)", IndexName, d, 100 * d.ElapsedTime / total.ElapsedTime);
+				Log.Debug ("###### {0}:    Gen'd Results {1,6} ({2:0.0}%)", IndexName, e, 100 * e.ElapsedTime / total.ElapsedTime);
+				Log.Debug ("###### {0}:   Reader cleanup {1,6} ({2:0.0}%)", IndexName, f, 100 * f.ElapsedTime / total.ElapsedTime);
+				Log.Debug ("###### {0}:            TOTAL {1,6}", IndexName, total);
+
+				Logger.Log.Debug ("###### {0}: Total query run in {1}", IndexName, total);
+			}
 
 		}
 
@@ -474,17 +544,16 @@ namespace Beagle.Daemon {
 
 		////////////////////////////////////////////////////////////////
 
-		static private void ScoreHits (Hashtable   hits_by_id,
+		static private void ScoreHits (Dictionary<int, Hit>   hits_by_id,
 					       IndexReader reader,
 					       ICollection term_list)
 		{
-			Stopwatch sw;
-			sw = new Stopwatch ();
-			sw.Start ();
-
 			LNS.Similarity similarity;
 			similarity = LNS.Similarity.GetDefault ();
 
+			TermDocs term_docs = reader.TermDocs ();
+			Hit hit;
+
 			foreach (Term term in term_list) {
 
 				double idf;
@@ -493,16 +562,13 @@ namespace Beagle.Daemon {
 				int hit_count;
 				hit_count = hits_by_id.Count;
 
-				TermDocs term_docs;
-				term_docs = reader.TermDocs (term);
+				term_docs.Seek (term);
 				while (term_docs.Next () && hit_count > 0) {
 					
 					int id;
 					id = term_docs.Doc ();
 
-					Hit hit;
-					hit = hits_by_id [id] as Hit;
-					if (hit != null) {
+					if (hits_by_id.TryGetValue (id, out hit)) {
 						double tf;
 						tf = similarity.Tf (term_docs.Freq ());
 						hit.Score += tf * idf;
@@ -511,38 +577,11 @@ namespace Beagle.Daemon {
 				}
 			}
 
-			sw.Stop ();
+			term_docs.Close ();
 		}
 
 		////////////////////////////////////////////////////////////////
 
-		// Lame iterator methods because we use two different ones
-		// depending on which algorithm we use.
-		private static IEnumerable IterateMatches (BetterBitArray primary_matches)
-		{
-			int j = primary_matches.Count;
-
-			// Walk across the matches backwards, since newer
-			// documents are more likely to be at the end of
-			// the index.
-			while (true) {
-				int i = primary_matches.GetPreviousTrueIndex (j);
-				if (i < 0)
-					yield break;
-				j = i-1; // This way we can't forget to adjust i
-
-				yield return i;
-			}
-		}
-
-		private static IEnumerable IterateOrderedMatches (int[] ordered_matches)
-		{
-			int i;
-
-			for (i = 0; i < ordered_matches.Length; i++)
-				yield return ordered_matches [i];
-		}
-
 		private class DocAndId {
 			public Document Doc;
 			public int Id;
@@ -553,6 +592,10 @@ namespace Beagle.Daemon {
 		// results.
 		//
 
+		// Two arrays we need for quickly creating lucene documents and check if they are valid
+		static string[] fields_timestamp_uri = { "Timestamp", "Uri" };
+		static string[] fields_uri = {"Uri"};
+
 		private static void GenerateQueryResults (IndexReader       primary_reader,
 							  LNS.IndexSearcher primary_searcher,
 							  LNS.IndexSearcher secondary_searcher,
@@ -564,8 +607,7 @@ namespace Beagle.Daemon {
 							  HitFilter         hit_filter,
 							  string            index_name)
 		{
-			TopScores top_docs = null;
-			ArrayList all_docs = null;
+			int num_hits;
 
 			if (Debug)
 				Logger.Log.Debug (">>> {0}: Initially handed {1} matches", index_name, primary_matches.TrueCount);
@@ -573,114 +615,248 @@ namespace Beagle.Daemon {
 			if (primary_matches.TrueCount <= max_results) {
 				if (Debug)
 					Logger.Log.Debug (">>> {0}: Initial count is within our limit of {1}", index_name, max_results);
-				all_docs = new ArrayList ();
+				num_hits = primary_matches.TrueCount;
 			} else {
 				if (Debug)
 					Logger.Log.Debug (">>> {0}: Number of hits is capped at {1}", index_name, max_results);
-				top_docs = new TopScores (max_results);
+				num_hits = max_results;
 			}
 
-			Stopwatch total, a, b, c, d;
+			Stopwatch total, d, e;
 			total = new Stopwatch ();
-			a = null;
-			b = new Stopwatch ();
-			c = new Stopwatch ();
 			d = new Stopwatch ();
+			e = new Stopwatch ();
 
 			total.Start ();
 
-			// There are two ways we can determine the max_results
-			// most recent items:  
-			//
-			// One is to instantiate Lucene documents for each of
-			// the document IDs in primary_matches.  This is a
-			// fairly expensive operation.
-			//
-			// The other is to walk through the list of all
-			// document IDs in descending time order.  This is
-			// a less expensive operation, but adds up over time
-			// on large data sets.
-			//
-			// We can walk about 2.5 docs for every Document we
-			// instantiate.  So what we'll do, if we have more
-			// matches than available hits, is walk (m * 1.25)
-			// docs to see if we can fill out the top 100 hits.
-			// If not, we'll fall back to creating documents
-			// for all of them.
+			ArrayList final_list_of_hits = null;
 
-			int[] ordered_matches = null;
-			if (primary_matches.TrueCount > max_results) {
-				a = new Stopwatch ();
-				a.Start ();
+			// This is used only for scoring
+			Dictionary<int, Hit> hits_by_id = new Dictionary<int, Hit> (num_hits);
 
-				TermDocs docs = primary_reader.TermDocs ();
-				TermEnum enumerator = primary_reader.Terms (new Term ("InvertedTimestamp", ""));
-				ordered_matches = new int [max_results];
-				int docs_found = 0;
-				int docs_walked = 0;
-				int max_docs = (int) (primary_matches.TrueCount * 1.25);
+			if (primary_matches.TrueCount > max_results)
+				final_list_of_hits = ScanRecentDocs (primary_reader,
+					primary_searcher,
+					secondary_searcher,
+					primary_matches,
+					hits_by_id,
+					max_results,
+					uri_filter,
+					index_name);
 
-				do {
-					Term term = enumerator.Term ();
-				
-					if (term.Field () != "InvertedTimestamp")
-						break;
+			if (final_list_of_hits == null)
+				final_list_of_hits = FindRecentResults (primary_reader,
+					primary_searcher,
+					secondary_searcher,
+					primary_matches,
+					hits_by_id,
+					max_results,
+					uri_filter,
+					index_name);
 
-					docs.Seek (enumerator);
+			d.Start ();
 
-					while (docs.Next ()
-					       && docs_found < ordered_matches.Length
-					       && docs_walked < max_docs) {
-						int doc_id = docs.Doc ();
+			ScoreHits (hits_by_id, primary_reader, query_term_list);
+			hits_by_id = null;
 
-						if (primary_matches.Get (doc_id)) {
-							ordered_matches [docs_found] = docs.Doc ();
-							docs_found++;
-						}
-				
-						docs_walked++;
+			d.Stop ();
+
+			if (Debug)
+				Log.Debug (">>> {0}: Scored hits in {1}", index_name, d);
+
+			e.Start ();
+
+			int total_number_of_matches = primary_matches.TrueCount;
+
+			// 25 hits seems to be the sweet spot: anything lower
+			// and serialization overhead gets us, higher takes
+			// longer to send out.
+			const int MAX_QUEUED_HITS = 25;
+			int sent_index = 0;
+
+			// Break up the hits into reasonably sized chunks for
+			// sending over the wire.  Also run our hit_filter, if
+			// we have one.
+			for (int i = 0; i < final_list_of_hits.Count; ++i) {
+				if (hit_filter != null) {
+					Hit hit = (Hit) final_list_of_hits [i];
+					if (! hit_filter (hit)) {
+						if (Debug)
+							Log.Debug ("Filtered out {0}", hit.Uri);
+						final_list_of_hits [i] = null;
+						total_number_of_matches --;
 					}
-				} while (enumerator.Next ()
-					 && docs_found < ordered_matches.Length
-					 && docs_walked < max_docs);
-
-				docs.Close ();
-
-				// We've found all the docs we can return in a subset!
-				// Fantastic, we've probably short circuited a slow search.
-				if (docs_found == max_results) {
-					all_docs = new ArrayList ();
-					top_docs = null;
-				} else {
-					// Bad luck! Not all docs found
-					// Start afresh - this time traversing all results
-					ordered_matches = null;
 				}
 
-				a.Stop ();
-				if (Debug) {
-					Log.Debug (">>> {0}: Walked {1} items, populated an enum with {2} items", index_name, docs_walked, docs_found, a);
-					
-					if (docs_found == max_results)
-						Log.Debug (">>> {0}: Successfully short circuited timestamp ordering!", index_name);
+				// Flush our hits
+				if (i > 0 && i % MAX_QUEUED_HITS == 0) {
+					result.Add (final_list_of_hits.GetRange (0, MAX_QUEUED_HITS));
+					final_list_of_hits.RemoveRange (0, MAX_QUEUED_HITS);
+					i -= MAX_QUEUED_HITS;
 				}
 			}
 
+			// Flush the remaining hits
+			result.Add (final_list_of_hits, total_number_of_matches);
+			final_list_of_hits = null;
+
+			e.Stop ();
+
+			if (Debug)
+				Log.Debug (">>> {0}: Hit filters executed and results sent in {1}", index_name, e);
+
+			total.Stop ();
+
+			if (Debug) {
+				Logger.Log.Debug (">>> {0}: GenerateQueryResults time statistics:", index_name);
+				//Logger.Log.Debug (">>> {0}:   Short circuit {1,6} ({2:0.0}%)", index_name, a == null ? "N/A" : a.ToString (), a == null ? 0.0 : 100 * a.ElapsedTime / total.ElapsedTime);
+				//Logger.Log.Debug (">>> {0}:     Create docs {1,6} ({2:0.0}%)", index_name, b, 100 * b.ElapsedTime / total.ElapsedTime);
+				//Logger.Log.Debug (">>> {0}:    Hit assembly {1,6} ({2:0.0}%)", index_name, c, 100 * c.ElapsedTime / total.ElapsedTime);
+				Logger.Log.Debug (">>> {0}:     Scored hits {1,6} ({2:0.0}%)", index_name, d, 100 * d.ElapsedTime / total.ElapsedTime);
+				Logger.Log.Debug (">>> {0}:    Results sent {1,6} ({2:0.0}%)", index_name, e, 100 * e.ElapsedTime / total.ElapsedTime);
+				Logger.Log.Debug (">>> {0}:           TOTAL {1,6}", index_name, total);
+			}
+		}
+
+		// There are two ways we can determine the max_results
+		// most recent items:  
+		//
+		// One is to instantiate Lucene documents for each of
+		// the document IDs in primary_matches.  This is a
+		// fairly expensive operation.
+		//
+		// The other is to walk through the list of all
+		// document IDs in descending time order.  This is
+		// a less expensive operation, but adds up over time
+		// on large data sets.
+		//
+		// We can walk about 2.5 docs for every Document we
+		// instantiate.  So what we'll do, if we have more
+		// matches than available hits, is walk (m * 1.25)
+		// docs to see if we can fill out the top 100 hits.
+		// If not, we'll fall back to creating documents
+		// for all of them.
+
+		private static ArrayList ScanRecentDocs (IndexReader	    primary_reader,
+						    LNS.IndexSearcher	    primary_searcher,
+						    LNS.IndexSearcher	    secondary_searcher,
+						    BetterBitArray	    primary_matches,
+						    Dictionary<int, Hit>    hits_by_id,
+						    int			    max_results,
+						    UriFilter		    uri_filter,
+						    string		    index_name)
+		{
+			Stopwatch a = new Stopwatch ();
+			a.Start ();
+
+			TermDocs docs = primary_reader.TermDocs ();
+			TermEnum enumerator = primary_reader.Terms (new Term ("InvertedTimestamp", String.Empty));
+			ArrayList results = new ArrayList (max_results);
+			int docs_found = 0;
+			int docs_walked = 0;
+			int max_docs = (int) (primary_matches.TrueCount * 1.25);
+
+			Term term;
+			TermDocs secondary_term_docs = null;
+			if (secondary_searcher != null)
+				secondary_term_docs = secondary_searcher.Reader.TermDocs ();
+
+			do {
+				term = enumerator.Term ();
+			
+				if (term.Field () != "InvertedTimestamp")
+					break;
+
+				docs.Seek (enumerator);
+
+				while (docs.Next ()
+				       && docs_found < max_results
+				       && docs_walked < max_docs) {
+					int doc_id = docs.Doc ();
+
+					if (primary_matches.Get (doc_id)) {
+						Document doc = primary_searcher.Doc (doc_id);
+						// If we have a UriFilter, apply it.
+						if (uri_filter != null) {
+							Uri uri;
+							uri = GetUriFromDocument (doc);
+							if (uri_filter (uri)) {
+								Hit hit = CreateHit (doc, secondary_searcher, secondary_term_docs);
+								hits_by_id [doc_id] = hit;
+								// Add the result, last modified first
+								results.Add (hit);
+								docs_found++;
+							}
+						}
+					}
+			
+					docs_walked++;
+				}
+			} while (enumerator.Next ()
+				 && docs_found < max_results
+				 && docs_walked < max_docs);
+
+			docs.Close ();
+			if (secondary_term_docs != null)
+				secondary_term_docs.Close ();
+
+			// If we've found all the docs we can return in a subset!
+			// Fantastic, we've probably short circuited a slow search.
+			if (docs_found != max_results) {
+				// Otherwise bad luck! Not all docs found
+				// Start afresh - this time traversing all results
+				results = null;
+			}
+
+			a.Stop ();
+			if (Debug) {
+				Log.Debug (">>> {0}: Walked {1} items, populated an enum with {2} items in {3}", index_name, docs_walked, docs_found, a);
+				
+				if (docs_found == max_results)
+					Log.Debug (">>> {0}: Successfully short circuited timestamp ordering!", index_name);
+			}
+
+			return results;
+		}
+
+		private static ArrayList   FindRecentResults (IndexReader	    primary_reader,
+							      LNS.IndexSearcher primary_searcher,
+							      LNS.IndexSearcher	    secondary_searcher,
+							      BetterBitArray	    primary_matches,
+							      Dictionary<int, Hit>  hits_by_id,
+							      int		    max_results,
+							      UriFilter		    uri_filter,
+							      string		    index_name)
+		{
+			Stopwatch b = new Stopwatch ();
 			b.Start ();
 			
 			int count = 0;
-			IEnumerable enumerable;
+			Document doc;
 
-			if (ordered_matches != null) {
-				enumerable = IterateOrderedMatches (ordered_matches);
-		       } else
-				enumerable = IterateMatches (primary_matches);
+			ArrayList all_docs = null;
+			TopScores top_docs = null;
+			TermDocs term_docs = null;
+
+			if (primary_matches.TrueCount > max_results)
+				top_docs = new TopScores (max_results);
+			else
+				all_docs = new ArrayList (primary_matches.TrueCount);
+
+			if (secondary_searcher != null)
+				term_docs = secondary_searcher.Reader.TermDocs ();
+
+			for (int match_index = primary_matches.Count; ; match_index --) {
+				// Walk across the matches backwards, since newer
+				// documents are more likely to be at the end of
+				// the index.
+				match_index = primary_matches.GetPreviousTrueIndex (match_index);
+				if (match_index < 0)
+					break;
 
-			foreach (int i in enumerable) {
 				count++;
 
-				Document doc;
-				doc = primary_searcher.Doc (i);
+				doc = primary_searcher.Doc (match_index, fields_timestamp_uri);
 
 				// Check the timestamp --- if we have already reached our
 				// limit, we might be able to reject it immediately.
@@ -700,157 +876,64 @@ namespace Beagle.Daemon {
 				if (uri_filter != null) {
 					Uri uri;
 					uri = GetUriFromDocument (doc);
-					if (! uri_filter (uri)) 
+					if (! uri_filter (uri))
 						continue;
 				}
 
-				DocAndId doc_and_id = new DocAndId ();
-				doc_and_id.Doc = doc;
-				doc_and_id.Id = i;
+				// Get the actual hit now
+				// doc was created with only 2 fields, so first get the complete lucene document for primary document
+				Hit hit = CreateHit (primary_searcher.Doc (match_index), secondary_searcher, term_docs);
+				hits_by_id [match_index] = hit;
 
 				// Add the document to the appropriate data structure.
 				// We use the timestamp_num as the score, so high
 				// scores correspond to more-recent timestamps.
 				if (all_docs != null)
-					all_docs.Add (doc_and_id);
+					all_docs.Add (hit);
 				else
-					top_docs.Add (timestamp_num, doc_and_id);
+					top_docs.Add (timestamp_num, hit);
 			}
 
-			if (Debug)
-				Log.Debug (">>> {0}: Processed roughly {1} documents", index_name, count);
+			if (term_docs != null)
+				term_docs.Close ();
 
 			b.Stop ();
 
-			c.Start ();
-
-			ICollection final_list_of_docs;
-			if (all_docs != null)
-				final_list_of_docs = all_docs;
-			else
-				final_list_of_docs = top_docs.TopScoringObjects;
-
-			ArrayList final_list_of_hits;
-			final_list_of_hits = new ArrayList (final_list_of_docs.Count);
-
-			// This is used only for scoring
-			Hashtable hits_by_id = null;
-			hits_by_id = new Hashtable ();
-
-			// If we aren't using the secondary index, the next step is
-			// very straightforward.
-			if (secondary_searcher == null) {
-
-				foreach (DocAndId doc_and_id in final_list_of_docs) {
-					Hit hit;
-					hit = DocumentToHit (doc_and_id.Doc);
-					hits_by_id [doc_and_id.Id] = hit;
-					final_list_of_hits.Add (hit);
-				}
+			if (Debug)
+				Log.Debug (">>> {0}: Instantiated and scanned {1} documents in {2}", index_name, count, b);
 
+			if (all_docs != null) {
+				// Sort results before sending
+				all_docs.Sort ();
+				return all_docs;
 			} else {
-
-				if (Debug)
-					Logger.Log.Debug (">>> {0}: Performing cross-index Hit reunification", index_name);
-
-				Hashtable hits_by_uri;
-				hits_by_uri = UriFu.NewHashtable ();
-
-				LuceneBitArray secondary_matches;
-				secondary_matches = new LuceneBitArray (secondary_searcher);
-
-				foreach (DocAndId doc_and_id in final_list_of_docs) {
-					Hit hit;
-					hit = DocumentToHit (doc_and_id.Doc);
-					hits_by_id [doc_and_id.Id] = hit;
-					hits_by_uri [hit.Uri] = hit;
-					secondary_matches.AddUri (hit.Uri);
-				}
-
-				secondary_matches.FlushUris ();
-				
-				// Attach all of our secondary properties
-				// to the hits
-				int j = 0;
-				while (true) {
-					int i = secondary_matches.GetNextTrueIndex (j);
-					if (i >= secondary_matches.Count)
-						break;
-					j = i+1;
-
-					Document secondary_doc;
-					secondary_doc = secondary_searcher.Doc (i);
-					
-					Uri uri;
-					uri = GetUriFromDocument (secondary_doc);
-
-					Hit hit;
-					hit = hits_by_uri [uri] as Hit;
-
-					AddPropertiesToHit (hit, secondary_doc, false);
-
-					final_list_of_hits.Add (hit);
-				}
-			}
-
-			ScoreHits (hits_by_id, primary_reader, query_term_list);
-
-			c.Stop ();
-
-			// If we used the TopScores object, we got our original
-			// list of documents sorted for us.  If not, sort the
-			// final list.
-			if (top_docs == null)
-				final_list_of_hits.Sort ();
-
-			d.Start ();
-
-			// If we have a hit_filter, use it now.
-			if (hit_filter != null) {
-				for (int i = 0; i < final_list_of_hits.Count; ++i) {
-					Hit hit;
-					hit = final_list_of_hits [i] as Hit;
-					if (! hit_filter (hit)) {
-						if (Debug)
-							Logger.Log.Debug ("Filtered out {0}", hit.Uri);
-						final_list_of_hits [i] = null;
-					}
-				}
-			}
-
-			// Before we broadcast a hit, we strip out any
-			// properties in the PrivateNamespace.  We
-			// manipulate the property ArrayList directory,
-			// which is pretty gross... but this is safe,
-			// since removing items will not change the sort
-			// order.
-			foreach (Hit hit in final_list_of_hits) {
-				if (hit == null)
-					continue;
-				int i = 0;
-				while (i < hit.Properties.Count) {
-					Property prop = hit.Properties [i] as Property;
-					if (prop.Key.StartsWith (Property.PrivateNamespace))
-						hit.Properties.RemoveAt (i);
-					else
-						++i;
-				}
-			}
-
-			result.Add (final_list_of_hits, primary_matches.TrueCount);
-
-			d.Stop ();
-			total.Stop ();
-
-			if (Debug) {
-				Logger.Log.Debug (">>> {0}: GenerateQueryResults time statistics:", index_name);
-				Logger.Log.Debug (">>> {0}: Short circuit {1,6} ({2:0.0}%)", index_name, a == null ? "N/A" : a.ToString (), a == null ? 0.0 : 100 * a.ElapsedTime / total.ElapsedTime);
-				Logger.Log.Debug (">>> {0}:    First pass {1,6} ({2:0.0}%)", index_name, b, 100 * b.ElapsedTime / total.ElapsedTime);
-				Logger.Log.Debug (">>> {0}:  Hit assembly {1,6} ({2:0.0}%)", index_name, c, 100 * c.ElapsedTime / total.ElapsedTime);
-				Logger.Log.Debug (">>> {0}:    Final pass {1,6} ({2:0.0}%)", index_name, d, 100 * d.ElapsedTime / total.ElapsedTime);
-				Logger.Log.Debug (">>> {0}:         TOTAL {1,6}", index_name, total);
+				return top_docs.TopScoringObjects;
 			}
 		}
 
+		private static Hit CreateHit ( Document primary_doc,
+					LNS.IndexSearcher secondary_searcher,
+					TermDocs term_docs)
+		{
+			Hit hit = DocumentToHit (primary_doc);
+
+			if (secondary_searcher == null)
+				return hit;
+
+			// Get the stringified version of the URI
+			// exactly as it comes out of the index.
+			Term term = new Term ("Uri", primary_doc.Get ("Uri"));
+			term_docs.Seek (term);
+
+			// Move to the first (and only) matching term doc
+			term_docs.Next ();
+			Document secondary_doc = secondary_searcher.Doc (term_docs.Doc ());
+
+			// If we are using the secondary index, now we need to
+			// merge the properties from the secondary index
+			AddPropertiesToHit (hit, secondary_doc, false);
+
+			return hit;
+		}
 	}
 }
